#!.venv/bin/python3
import pyAgrum as gum
import psycopg2
import logging
from psycopg2.extras import LoggingConnection
import sys
import argparse
import time


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

def is_var_indicator(v):
    return not is_var(v)

def is_var(v):
    return not v.lower().startswith('i')


parser = argparse.ArgumentParser()
parser.add_argument("BNFile", help="Bayesian network file")
parser.add_argument("DBSize", help="size of the database")
args = parser.parse_args()

# inputs
bn_file = args.BNFile
db_size = int(args.DBSize)
out_file = "mcc.csv"
table_prefix = "test"

# loading the BN file
bn = gum.loadBN(bn_file)
var_names = sorted(list(bn.names()))

# generating the CSV file
g=gum.BNDatabaseGenerator(bn)
g.drawSamples(db_size)
g.toCSV(out_file)

# PG connection
conn = psycopg2.connect(database="mcc", 
                        user='postgres', password='postgres',  
                        host='127.0.0.1', port='5432' )#, connection_factory=LoggingConnection)
# conn.initialize(logger)
conn.autocommit = True
cursor = conn.cursor()


#### CREATION OF D #####

# deleting the existing table
deletion_sql = "DROP TABLE IF EXISTS {};".format(table_prefix)
cursor.execute(deletion_sql)

# creating the table to store the data without missing values
col_def = map(lambda name : "{} int NOT NULL".format(name), var_names)
creation_sql = "CREATE TABLE {}(block SERIAL PRIMARY KEY, {});".format(table_prefix, ",".join(col_def))
cursor.execute(creation_sql) 

# loading the CSV data to the table
with open(out_file, "r") as file:
    next(file) # skip the header's line
    cursor.copy_from(file, table_prefix, sep=",", null="", columns=[v.lower() for v in var_names])

#### CREATION OF D STAR ##### 

missing_selects = []
for v in var_names:
    if is_var(v) and ("i{}".format(v) in var_names or "I{}".format(v) in var_names):
            missing_selects.append("case when i{}=1 then NULL ELSE {} END".format(v,v))
    elif is_var(v):
        missing_selects.append(v)

missing_data_table = "{}_star".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(missing_data_table)
cursor.execute(deletion_sql)
missing_selects_text = ",".join(missing_selects)
missing_data_sql = "CREATE TABLE {} AS SELECT block, {}, md5(CAST(({}) AS text)) AS superblock FROM {}".format(missing_data_table, missing_selects_text, missing_selects_text, table_prefix)
cursor.execute(missing_data_sql)

#### CREATION OF BID #####

# superblock definitions with full tuples
superblock_tmp_table = "{}_sb_tmp".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(superblock_tmp_table)
cursor.execute(deletion_sql)

vars = list(filter(is_var, var_names))
col_def = map(lambda name : "{} int NOT NULL".format(name), vars)
creation_sql = "CREATE TABLE {}(superblock text, {}, prob float NOT NULL);".format(superblock_tmp_table, ",".join(col_def))
cursor.execute(creation_sql) 

cursor.execute("SELECT DISTINCT {}, superblock FROM {}".format(",".join(vars), missing_data_table))
ie = gum.LazyPropagation(bn)

insert_sql = "INSERT INTO {}({}, superblock, prob) VALUES({},%s,%s)".format(superblock_tmp_table, ",".join(vars), ",".join(["%s" for _ in vars]))

def instantiation_to_list(inst, row, vars):
    l = []
    i = 0
    inst_vars = list(map(lambda v : v.name(), inst.variablesSequence())) if inst is not None else []
    for v in vars:
        if v in inst_vars:
            l.append(inst[v])
        else:
            l.append(row[i])
        i+=1
    return l

for row in cursor.fetchall():
    null_vars = []
    null_pos  = []
    i = 0
    for v in vars:
        if row[i] is None:
            null_vars.append(v)
            null_pos.append(i)
        else:
            ie.addEvidence(v, row[i])
        i+=1
    ie.addJointTarget(set(null_vars))
    if len(null_vars) > 0:
        potentiel = ie.jointPosterior(set(null_vars)) if len(vars) > len(null_vars) else ie.evidenceJointImpact(vars,{})
        for inst in potentiel.loopIn():
            inserted_row = instantiation_to_list(inst, row, vars)
            inserted_row.append(row[-1])
            inserted_row.append(potentiel.get(inst))
            cursor.execute(insert_sql, inserted_row)
    else:
        inserted_row = instantiation_to_list(None, row, vars)
        inserted_row.append(row[-1])
        inserted_row.append(1)
        cursor.execute(insert_sql, inserted_row)
    ie.eraseAllJointTargets()
    ie.eraseAllEvidence()

# tuples table creation
tuple_table = "{}_tuple".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(tuple_table)
cursor.execute(deletion_sql)

cursor.execute("CREATE TABLE {} AS SELECT DISTINCT ROW_NUMBER() OVER (ORDER BY (SELECT 1)) AS tuple, * FROM (SELECT DISTINCT {} FROM {}) as t".format(tuple_table, ",".join(vars), superblock_tmp_table))

# superblocks table creation with tuple identifiers
superblock_table = "{}_sb".format(table_prefix)
deletion_sql = "DROP TABLE IF EXISTS {};".format(superblock_table)
cursor.execute(deletion_sql)

sb_tuple_preds = []
for v in vars:
    pred = "sb.{} = t.{}".format(v, v)
    sb_tuple_preds.append(pred)

cursor.execute("CREATE TABLE {} AS SELECT sb.superblock, t.tuple, sb.prob FROM {} AS sb, {} AS t WHERE {}".format(superblock_table, superblock_tmp_table, tuple_table, " AND ".join(sb_tuple_preds)))
    
deletion_sql = "DROP TABLE IF EXISTS {};".format(superblock_tmp_table)
cursor.execute(deletion_sql)



### Computation of the MCC

tuple_label_table = "{}_tuple_label".format(table_prefix)
block_label_table = "{}_block_label".format(table_prefix)
graph_table = "{}_graph".format(table_prefix)

def tuple_joint_probability(tuple):
    i = bn.completeInstantiation()
    n = 0
    p = ie.evidenceJointImpact(vars,{})
    for v in i.variablesSequence():
        if n < len(tuple):
            i.chgVal(v, tuple[n])
        else:
            i.chgVal(v, 1)
        n += 1
    return p.get(i)

def tuple_occurence_weight(tuple, occurence):
    return 2*db_size*tuple_joint_probability(tuple) - 2*occurence +1

                 
def initialize_labels():
    deletion_sql = "DROP TABLE IF EXISTS {};".format(tuple_label_table)
    cursor.execute(deletion_sql)
    cursor.execute("CREATE TABLE {} (tuple INTEGER, occurence INTEGER, label float, weight float, PRIMARY KEY (tuple, occurence)) ".format(tuple_label_table))

    insert_sql = "INSERT INTO {}(tuple, occurence, label, weight) VALUES(%s,%s,%s,%s)".format(tuple_label_table)
    cursor.execute("SELECT * FROM {}".format(tuple_table))

    for row in cursor.fetchall():
        inserted_row = []
        tuple_id = row[0]
        tuple = row[1:]
        inserted_row.append(tuple_id)
        inserted_row.append(1)
        inserted_row.append(0)
        inserted_row.append(tuple_occurence_weight(tuple, 1))
        # inserted_row.append(tuple_id + 2*db_size)
        cursor.execute(insert_sql, inserted_row)
        
    deletion_sql = "DROP TABLE IF EXISTS {};".format(block_label_table)
    cursor.execute(deletion_sql)
    cursor.execute("CREATE TABLE {} AS SELECT m.block, MAX(tl.weight) as label from {} as m, {} as s, {} as tl where m.superblock = s.superblock and s.tuple = tl.tuple GROUP BY m.block; ".format(block_label_table, missing_data_table, superblock_table, tuple_label_table))
    cursor.execute("CREATE UNIQUE INDEX ON {} (block)".format(block_label_table))

def initialize_graph():
    deletion_sql = "DROP TABLE IF EXISTS {};".format(graph_table)
    cursor.execute(deletion_sql)
    cursor.execute("CREATE TABLE {} (block INTEGER, tuple INTEGER, occurence INTEGER, eq BOOLEAN, matching BOOLEAN DEFAULT false, PRIMARY KEY(block, tuple, occurence)) ".format(graph_table))
    cursor.execute("INSERT INTO {}(block, tuple, occurence, eq)  SELECT m.block, s.tuple, 1 AS occurence, false AS eq FROM {} as m, {} as s where m.superblock = s.superblock".format(graph_table, missing_data_table, superblock_table))
    cursor.execute("UPDATE {} SET eq = true WHERE (tuple, block) IN (SELECT tl.tuple, bl.block FROM {} AS tl, {} AS bl WHERE tl.weight = bl.label);".format(graph_table, tuple_label_table, block_label_table))
    cursor.execute("CREATE INDEX ON {} (block)".format(graph_table))
    cursor.execute("CREATE INDEX ON {} (tuple,occurence)".format(graph_table))
    cursor.execute("CREATE INDEX ON {} (eq)".format(graph_table))
    cursor.execute("CREATE INDEX ON {} (matching)".format(graph_table))

# def initialize_matching():
#     deletion_sql = "DROP TABLE IF EXISTS {};".format(matching_table)
#     cursor.execute(deletion_sql)
#     cursor.execute("CREATE TABLE {} (block INTEGER, tuple INTEGER, occurence INTEGER) ".format(matching_table))

# return a block free vertex in the equigraph, if it exists
def pick_free_block():
    cursor.execute("SELECT DISTINCT block FROM {} WHERE block NOT IN (SELECT block FROM  {} WHERE matching = true)".format(graph_table, graph_table))
    res = cursor.fetchone()
    return res[0] if res is not None else None

neigh_time = 0
# return the (tuple, occ) vertices that are the neighbors in the equigraph of at least one block in a set
def get_neighbors(blocks):
    start = time.monotonic()
    neighbors = set()
    cursor.execute("SELECT DISTINCT tuple, occurence FROM {} WHERE eq = true AND block IN %s".format(graph_table), [tuple(blocks)])
    for row in cursor.fetchall():
        neighbors.add(row)

    global neigh_time
    neigh_time += time.monotonic() - start

    return neighbors

matched_neigh_time = 0

def get_matching_block(tuple, occurence):
    start = time.monotonic()
    cursor.execute("SELECT block FROM {} WHERE tuple = {} AND occurence = {} AND matching = true".format(graph_table, tuple, occurence))
    res = cursor.fetchone()

    global matched_neigh_time
    matched_neigh_time += time.monotonic() - start

    return res[0] if res is not None else None

matching_time = 0

def augmenting_path_insertion(u, y, prec):
    start = time.monotonic()
    is_edge_matched = False
    while y != u:
        if is_edge_matched :
            cursor.execute("UPDATE {} SET matching = false WHERE block = %s AND tuple = %s AND occurence = %s".format(graph_table), [y, prec[y][0], prec[y][1]])
        else:
            cursor.execute("UPDATE {} SET matching = true WHERE block = %s AND tuple = %s AND occurence = %s".format(graph_table), [prec[y], y[0], y[1]])
        y = prec[y]
        is_edge_matched = not is_edge_matched
    global matching_time
    matching_time += time.monotonic() - start

lookup_time = 0
update_time1 = 0
update_time2 = 0
update_time3 = 0
label_update_nb = 0
edges_nb = 0
max_S = 0
        
def update_labels(S, T):
    start = time.monotonic()
    global label_update_nb
    label_update_nb += 1

    # first query for the edges not in the equigraph, but stored (implies occurence = 1)
    # second query for the edges not in the equigraph and not stored (not considered so far)
    cursor.execute("""
    SELECT g.block, g.tuple, g.occurence, tl.weight, bl.label, tl.label, true AS existing
    FROM {} AS g,  {} AS bl, {} AS tl
    WHERE tl.tuple = g.tuple AND tl.occurence = g.occurence AND bl.block = g.block
    AND g.eq = false AND g.block IN %s
    UNION
    SELECT g.block, g.tuple, g.occurence + 1, tl.weight - 2, bl.label, COALESCE(tln.label,0), false AS existing
    FROM {} AS g, {} AS bl, {} AS tl LEFT OUTER JOIN {} AS tln ON (tl.tuple = tln.tuple AND tl.occurence + 1 = tln.occurence)
    WHERE tl.tuple = g.tuple AND tl.occurence = g.occurence AND bl.block = g.block
    AND g.eq = true AND g.occurence < %s AND g.block IN %s 
    """.format(graph_table, block_label_table, tuple_label_table, graph_table, block_label_table, tuple_label_table, tuple_label_table), [tuple(S), db_size, tuple(S)])

    alpha = sys.float_info.max
    edges_to_add = []
    for row in cursor.fetchall():
        # print(row)
        if (row[1], row[2]) in T:
            # skip the edge not considered so far whose tuple is in T   
            continue
        a = row[4] + row[5] - row[3]
        if a < alpha :
            alpha = a
            edges_to_add = []
        if a == alpha:
            edges_to_add.append(row)

    mt = time.monotonic()
    global lookup_time
    lookup_time += mt - start
    mt2=mt
    
    cursor.execute("UPDATE {} SET label = label - %s WHERE block in %s".format(block_label_table), [alpha, tuple(S)])
    if len(T) > 0:
        cursor.execute("UPDATE {} SET label = label + %s WHERE (tuple, occurence) in %s".format(tuple_label_table), [alpha, tuple(T)])
        # remove from the equigraph the edges from T and a block not in S
        mt2 = time.monotonic()
        global update_time1
        update_time1 += mt2 - mt
        cursor.execute("DELETE FROM {} WHERE eq = True AND occurence > 1 AND (tuple, occurence) in %s AND block NOT IN %s".format(graph_table), [tuple(T), tuple(S)])
        cursor.execute("UPDATE {} SET eq = False WHERE occurence = 1 AND (tuple, occurence) in %s AND block NOT IN %s".format(graph_table), [tuple(T), tuple(S)])

    mt3 = time.monotonic()
    global update_time2
    update_time2 += mt3 - mt2
    global edges_nb
    edges_nb += len(edges_to_add)
    global max_S
    max_S = max(len(S), max_S)

    for edge in edges_to_add:
        # if the edge exists in the graph already
        if edge[6] :
            cursor.execute("UPDATE {} SET eq = true WHERE block = %s AND tuple = %s AND occurence = %s".format(graph_table), [edge[0], edge[1], edge[2]])
        else:
            cursor.execute("INSERT  INTO {} (tuple, occurence, label, weight) VALUES (%s, %s, %s, %s) ON CONFLICT DO NOTHING".format(tuple_label_table), [edge[1], edge[2], edge[5], edge[3]])
            cursor.execute("INSERT  INTO {} (block, tuple, occurence, eq) VALUES (%s, %s, %s, true)".format(graph_table), [edge[0], edge[1], edge[2]])

    global update_time3
    update_time3 += time.monotonic() - mt3

    return list(map(lambda e: [e[0], (e[1],e[2])] ,edges_to_add))

exploring_step_nb = 0

def hungarian_step(u, S, T, S_neigh, prec):
    # print("u={}".format(u))
    while True :
        remaining_neigh = S_neigh - T
        if len(remaining_neigh) > 0:
            global exploring_step_nb
            exploring_step_nb += 1
            y = next(iter(remaining_neigh))
            if len(S) == 1:
                prec[y] = u
            z = get_matching_block(y[0], y[1])
            if z is None :
                augmenting_path_insertion(u, y, prec)
                # print("augmenting path from {} to {} with S={} and T={} \n".format(u, y, S, T))
                return
            else:
                S.add(z)
                T.add(y)
                prec[z] = y
                for nz in get_neighbors({z}):
                    S_neigh.add(nz)
                    if nz != y and nz not in prec :
                        prec[nz] = z
                # print("matched y={} by z={} with S={}, T={} and neigh={}".format(y,z,S,T, S_neigh))
        else:
            new_edges = update_labels(S, T)
            for e in new_edges:
                S_neigh.add(e[1])
                if e[1] not in prec:
                    prec[e[1]] = e[0]


def hungarian_algorithm():
    for i in range(0, db_size):
        u = pick_free_block()
        S = {u}
        T = set()
        S_neigh = get_neighbors(S)
        prec = dict()

        hungarian_step(u, S, T, S_neigh, prec)
            
initialize_labels()
initialize_graph()


input("starting hungarian algorithm")
start = time.monotonic()
hungarian_algorithm()
print(f"exploring steps : {exploring_step_nb}")
print(f"neigh time in {neigh_time} s")
print(f"matched neigh time in {matched_neigh_time} s")
print(f"label update nb : {label_update_nb}")
print(f"lookup time in {lookup_time} s")
print(f"update time 1 in {update_time1} s")
print(f"edges nb : {edges_nb}")
print(f"max S : {max_S}")
print(f"update time 2 in {update_time2} s")
print(f"update time 3 in {update_time3} s")
print(f"matching time in {matching_time} s")
print(f"total  time in {time.monotonic() - start} s")

###### Imputation

# import numpy as np
# import pandas as pd
# from sklearn.impute import KNNImputer

# cursor.execute("SELECT * FROM {}".format(missing_data_table))
# tuples_list = cursor.fetchall()
# df = pd.DataFrame(tuples_list)

# imputer = KNNImputer(n_neighbors=10)
# impute = imputer.fit_transform(df)
# print(impute)
